"""convenience functions for prediction coming soon.
For now, please use `create_prediction_pipeline` to get a `PredictionPipeline`
and then `PredictionPipeline.predict_sample(sample)`
e..g load samples with core.io.load_sample_for_model()
"""

# TODO: add convenience functions for predictions
